/* SPDX-License-Identifier: GPL-2.0 */
/* Copyright (C) 2004  PaX Team <pageexec@freemail.hu> */

#ifndef _ASM_X86_KERNEXEC_H
#define _ASM_X86_KERNEXEC_H

#ifndef __ASSEMBLY__
#include <linux/compiler.h>
#include <linux/preempt.h>
#include <asm/bug.h>
#include <asm/special_insns.h>

#define  __HAVE_ARCH_PAX_OPEN_KERNEL
#define  __HAVE_ARCH_PAX_CLOSE_KERNEL

#ifdef CONFIG_PAX_KERNEXEC
static __always_inline unsigned long native_pax_open_kernel(void)
{
	unsigned long cr0;

	preempt_disable();
	barrier();
	// TODO: disable CR4.CET
	cr0 = read_cr0() ^ X86_CR0_WP;
	BUG_ON(cr0 & X86_CR0_WP);
	write_cr0(cr0);
	barrier();
	return cr0 ^ X86_CR0_WP;
}

static __always_inline unsigned long native_pax_close_kernel(void)
{
	unsigned long cr0;

	barrier();
	cr0 = read_cr0() ^ X86_CR0_WP;
	BUG_ON(!(cr0 & X86_CR0_WP));
	write_cr0(cr0);
	// TODO: enable CR4.CET
	barrier();
	preempt_enable_no_resched();
	return cr0 ^ X86_CR0_WP;
}
#else
static inline unsigned long native_pax_open_kernel(void) { return 0; }
static inline unsigned long native_pax_close_kernel(void) { return 0; }
#endif

#ifdef CONFIG_PARAVIRT_XXL
#include <asm/paravirt.h>
#else  /* !CONFIG_PARAVIRT_XXL */
static __always_inline unsigned long pax_open_kernel(void) __opens(kernel);
static __always_inline unsigned long pax_open_kernel(void)
{
	__open(kernel);
	return native_pax_open_kernel();
}

static __always_inline unsigned long pax_close_kernel(void) __closes(kernel);
static __always_inline unsigned long pax_close_kernel(void)
{
	__close(kernel);
	return native_pax_close_kernel();
}
#endif /* CONFIG_PARAVIRT_XXL */
#endif /* __ASSEMBLY__ */

#endif /* _ASM_X86_KERNEXEC_H */
