/* SPDX-License-Identifier: GPL-2.0 */
#ifndef _LINUX_LINKAGE_H
#define _LINUX_LINKAGE_H

#include <linux/compiler_types.h>
#include <linux/stringify.h>
#include <linux/export.h>
#include <asm/linkage.h>
#include <asm/bitsperlong.h>

/* Some toolchains use other characters (e.g. '`') to mark new line in macro */
#ifndef ASM_NL
#define ASM_NL		 ;
#endif

#ifdef __cplusplus
#define CPP_ASMLINKAGE extern "C"
#else
#define CPP_ASMLINKAGE
#endif

#ifndef asmlinkage
#define asmlinkage CPP_ASMLINKAGE
#endif

#ifndef cond_syscall
# ifdef CONFIG_PAX_RAP_HASH
#  define rap_cond_syscall(x)				\
	".weak " __stringify(rap_##x) "\n\t"		\
	".set  " __stringify(rap_##x) ","		\
		 __stringify(rap_sys_ni_syscall)
# else
#  define rap_cond_syscall(x)
# endif
#define cond_syscall(x)	asm(				\
	rap_cond_syscall(x)				\
	".weak " __stringify(x) "\n\t"			\
	".set  " __stringify(x) ","			\
		 __stringify(sys_ni_syscall))
#endif

#ifndef SYSCALL_ALIAS
#ifdef CONFIG_PAX_RAP_HASH
#define RAP_SYSCALL_ALIAS(alias, name)			\
	".globl " __stringify(rap_##alias) "\n\t"	\
	".set   " __stringify(rap_##alias) ","		\
		  __stringify(rap_##name)
#else
#define RAP_SYSCALL_ALIAS(alias, name)
#endif
#define SYSCALL_ALIAS(alias, name) asm(			\
	RAP_SYSCALL_ALIAS(alias, name)			\
	".globl " __stringify(alias) "\n\t"		\
	".set   " __stringify(alias) ","		\
		  __stringify(name))
#endif

#define __page_aligned_data	__section(.data..page_aligned) __aligned(PAGE_SIZE)
#define __page_aligned_rodata	__read_only __aligned(PAGE_SIZE)
#define __page_aligned_bss	__section(.bss..page_aligned) __aligned(PAGE_SIZE)

/*
 * For assembly routines.
 *
 * Note when using these that you must specify the appropriate
 * alignment directives yourself
 */
#define __PAGE_ALIGNED_DATA	.section ".data..page_aligned", "aw"
#define __PAGE_ALIGNED_BSS	.section ".bss..page_aligned", "aw"

/*
 * This is used by architectures to keep arguments on the stack
 * untouched by the compiler by keeping them live until the end.
 * The argument stack may be owned by the assembly-language
 * caller, not the callee, and gcc doesn't always understand
 * that.
 *
 * We have the return value, and a maximum of six arguments.
 *
 * This should always be followed by a "return ret" for the
 * protection to work (ie no more work that the compiler might
 * end up needing stack temporaries for).
 */
/* Assembly files may be compiled with -traditional .. */
#ifndef __ASSEMBLY__
#ifndef asmlinkage_protect
# define asmlinkage_protect(n, ret, args...)	do { } while (0)
#endif
#endif

#ifndef __ALIGN
#define __ALIGN		.align 4,0x90
#define __ALIGN_STR	".align 4,0x90"
#endif

#ifdef CONFIG_PAX_RAP_HASH
# if BITS_PER_LONG == 64
#  define RAP_RET_HASH_OFFSET		16
#  define RAP_HASH_LEN			8
#  define __ASM_RAP_HASH_PAD		.skip (16 - ((2 + RAP_HASH_LEN) % 16)) % 16, 0xcc
#  define __ASM_RAP_HASH(hash)		__ASM_RAP_HASH_PAD ASM_NL .byte 0x48, 0xb8 ASM_NL .quad hash
#  define __ASM_RAP_RET_HASH(hash)	.byte 0x48, 0xb8 ASM_NL .quad hash
# elif BITS_PER_LONG == 32
#  define RAP_RET_HASH_OFFSET		12
#  define RAP_HASH_LEN			4
#  define __ASM_RAP_HASH_PAD		.skip (16 - ((1 + RAP_HASH_LEN) % 16)) % 16, 0xcc
#  define __ASM_RAP_HASH(hash)		__ASM_RAP_HASH_PAD ASM_NL .byte 0xb8 ASM_NL .long hash
#  define __ASM_RAP_RET_HASH(hash)	.byte 0xb8 ASM_NL .long hash
# else
#  error incompatible BITS_PER_LONG
# endif
#endif

#ifdef __ASSEMBLY__

#ifndef LINKER_SCRIPT
#define ALIGN __ALIGN
#define ALIGN_STR __ALIGN_STR

#ifndef GLOBAL
#define GLOBAL(name) \
	.globl name ASM_NL \
	name:
#endif

#ifndef ENTRY
#define __ENTRY(name, rap_hash) \
	.globl name ASM_NL \
	ALIGN ASM_NL \
	rap_hash \
	name:

#define ENTRY(name) __ENTRY(name,)

#endif

#endif /* LINKER_SCRIPT */

#ifndef WEAK
#define __WEAK(name, rap_hash) \
	.weak name ASM_NL \
	ALIGN ASM_NL \
	rap_hash \
	name:

#define WEAK(name) __WEAK(name, )
#endif

#ifdef CONFIG_PAX_RAP_HASH
# define RAP_ENTRY(name) __ENTRY(name, __ASM_RAP_HASH(__rap_hash_call_##name) ASM_NL)
# define RAP_WEAK(name) __WEAK(name, __ASM_RAP_HASH(__rap_hash_call_##name) ASM_NL)
#else
# define RAP_ENTRY(name) ENTRY(name)
# define RAP_WEAK(name) WEAK(name)
#endif

#ifndef END
#define END(name) \
	.size name, .-name
#endif

/* If symbol 'name' is treated as a subroutine (gets called, and returns)
 * then please use ENDPROC to mark 'name' as STT_FUNC for the benefit of
 * static analysis tools such as stack depth analyzer.
 */
#ifndef ENDPROC
#define ENDPROC(name) \
	.type name, @function ASM_NL \
	END(name)
#endif

#else /* !__ASSEMBLY__*/
#ifdef CONFIG_PAX_RAP_HASH
#define RAP_HASH(name) __stringify(__ASM_RAP_HASH(__rap_hash_call_##name))
#define RAP_RET_HASH(name) __stringify(__ASM_RAP_HASH(__rap_hash_ret_##name))
#else
#define RAP_HASH(name)
#define RAP_RET_HASH(name)
#endif
#endif

#endif
