/*
 * Copyright 2019-2020 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * work around a gcc bug affecting gcc versions 4.7-7 that can result in uninitialized padding bytes in local variables
 */

#include "utilities.h"

#if BUILDING_GCC_VERSION >= 4007 && BUILDING_GCC_VERSION < 8000
static gimple_stmt_ptr barrier(tree var)
{
	tree input;
#if BUILDING_GCC_VERSION <= 4007
	VEC(tree, gc) *inputs = NULL;
#else
	vec<tree, va_gc> *inputs = NULL;
#endif

	input = build_tree_list(NULL_TREE, build_const_char_string(2, "g"));
	input = chainon(NULL_TREE, build_tree_list(input, var));
#if BUILDING_GCC_VERSION <= 4007
	VEC_safe_push(tree, gc, inputs, input);
#else
	vec_safe_push(inputs, input);
#endif

	return gimple_build_asm_vec("", inputs, NULL, NULL, NULL);
}

unsigned int fix_esra_execute(void)
{
	basic_block bb;

	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt, barrier_stmt;
			tree lhs, ref;

			stmt = gsi_stmt(gsi);
			if (!gimple_assign_copy_p(stmt))
				continue;

			if (gimple_clobber_p(stmt))
				continue;

			if (TREE_CODE(gimple_assign_rhs1(stmt)) != CONSTRUCTOR)
				continue;

			lhs = gimple_assign_lhs(stmt);
			if (DECL_P(lhs) && DECL_REGISTER(lhs))
				continue;

			ref = get_inner_arg(lhs);
			if (TREE_CODE(ref) == MEM_REF)
				ref = TREE_OPERAND(ref, 0);
			if (TREE_CODE(ref) != SSA_NAME)
				ref = build_unary_op(gimple_location(stmt), ADDR_EXPR, ref, false);
			barrier_stmt = barrier(ref);
//			gimple_asm_set_volatile(as_a_gasm(barrier_stmt), true);
			gimple_set_location(barrier_stmt, gimple_location(stmt));
			gimple_set_block(barrier_stmt, gimple_block(stmt));
			gsi_insert_after(&gsi, barrier_stmt, GSI_NEW_STMT);
		}
	}

	return 0;
}

#define PASS_NAME fix_esra
#define NO_GATE
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_remove_unused_locals | TODO_update_ssa | TODO_verify_flow | TODO_update_address_taken
#include "gcc-generate-gimple-pass.h"
#endif
