/*
 * Copyright 2020 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * allow specifying some ELF section flags not supported by gcc's section attr
 */

#include "utilities.h"

static tree handle_section_type_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	tree decl;
	int i;

	decl = *node;
	if (!VAR_P(decl)) {
		error_at(DECL_SOURCE_LOCATION(decl), "attribute %qE applies to variables only", name);
		return NULL_TREE;
	}

	if (TREE_CODE(TREE_VALUE(args)) != STRING_CST) {
		error_at(DECL_SOURCE_LOCATION(decl), "attribute %qE requires a string argument", name);
		return NULL_TREE;
	}

	args = TREE_VALUE(args);
	for (i = 0; i < TREE_STRING_LENGTH(args); i++) {
		switch (TREE_STRING_POINTER(args)[i]) {
		default:
			error_at(DECL_SOURCE_LOCATION(decl), "unknown value %c provided to attribute %qE", TREE_STRING_POINTER(args)[i], name);
			return NULL_TREE;

		case 0:
			break;

		case '1':
		case '2':
		case '4':
		case '8':
			flags |= TREE_STRING_POINTER(args)[i] - '0';
			break;

		case 'M':
			flags |= SECTION_MERGE;
			break;

		case 'n':
			flags &= ~SECTION_NOTYPE;
			break;

		case 'S':
			flags |= SECTION_STRINGS;
			break;
		}
	}

	*no_add_attrs = false;
	return NULL_TREE;
}

struct attribute_spec section_type_flags_attr = ATTRIBUTE_SPEC_INIT("section_type", 1, 1, true, false, false, handle_section_type_attribute, false, NULL);

static unsigned int (*old_section_type_flags)(tree decl, const char *name, int reloc);

static unsigned int utilities_section_type_flags(tree decl, const char *name, int reloc)
{
	tree section_type;
	unsigned int flags;

	flags = old_section_type_flags(decl, name, reloc);
	if (decl == NULL_TREE)
		return flags;

	section_type = lookup_attribute("section_type", DECL_ATTRIBUTES(decl));
	if (section_type != NULL_TREE) {
		int i;

		section_type = TREE_VALUE(TREE_VALUE(section_type));
		for (i = 0; i < TREE_STRING_LENGTH(section_type); i++) {
			switch (TREE_STRING_POINTER(section_type)[i]) {
			default:
				debug_tree(decl);
				gcc_unreachable();

			case 0:
				break;

			case '1':
			case '2':
			case '4':
			case '8':
				flags |= TREE_STRING_POINTER(section_type)[i] - '0';
				break;

			case 'M':
				flags |= SECTION_MERGE;
				break;

			case 'n':
				flags &= ~SECTION_NOTYPE;
				break;

			case 'S':
				flags |= SECTION_STRINGS;
				break;
			}
		}
	}

	return flags;
}

void section_type_start_unit(void __unused *gcc_data, void __unused *user_data)
{
	old_section_type_flags = targetm.section_type_flags;
	targetm.section_type_flags = utilities_section_type_flags;
}
